#include "main.h"
#include "raster.h"
#include "geometry.h"
#include "bitmap.h"
#include <stdio.h>

//---------------------------------------------------------------------------
//    
void Draw_Scene_Solid()
{
	//         
	Init_Raster_Frame();
	//      ( )
	Plane_struct plane(0,0,-1,-MINIMUM_Z);
	//     
	Line_struct left(-1,0,0),right(1,0,RWidth);
	Line_struct top(0,-1,0),bottom(0,1,RHeight);
	//   
	for(long i=0;i<Number_Faces;i++)
	{
		const Face_struct &face=Faces[i];
		//    
		Circle_List_Template<Vertex3D_struct> Poly3D;
		Poly3D.append(new Vertex3D_struct(Vertices3D[face.vertex[0]]));
		Poly3D.append(new Vertex3D_struct(Vertices3D[face.vertex[1]]));
		Poly3D.append(new Vertex3D_struct(Vertices3D[face.vertex[2]]));
		//    
		if (TruncPolygon3D(Poly3D,plane))
		{
			//    
			Circle_List_Template<Vertex2D_struct> Poly2D;
			ProjectPolygon(Poly3D,Poly2D);

			//    
			if (	TruncPolygon2D(Poly2D,left)	&&	TruncPolygon2D(Poly2D,right)
				&&	TruncPolygon2D(Poly2D,top)	&&	TruncPolygon2D(Poly2D,bottom)	)
			{
				//  
				Raster_Polygon(Poly2D,face.color_id);
			}

			//     
			Poly2D.clear();
		}

		//     
		Poly3D.clear();
	}
}

//---------------------------------------------------------------------------
//     
void Draw_Scene_Wire()
{
	//         
	Init_Raster_Frame();
	//      ( )
	Plane_struct plane(0,0,-1,-MINIMUM_Z);
	//     
	Line_struct left(-1,0,0),right(1,0,RWidth);
	Line_struct top(0,-1,0),bottom(0,1,RHeight);
	//   
	for(long i=0;i<Number_Edges;i++)
	{
		const Edge_struct &edge=Edges[i];
		//  
		Vertex3D_struct p1=Vertices3D[edge.vertex[0]],p2=Vertices3D[edge.vertex[1]];
		//    
		if (TruncSegment3D(p1,p2,plane))
		{
			//    
			Vertex2D_struct v1,v2;
			Project(p1,v1);
			Project(p2,v2);

			//    
			if (	TruncSegment2D(v1,v2,left)	&&	TruncSegment2D(v1,v2,right)
				&&	TruncSegment2D(v1,v2,top)	&&	TruncSegment2D(v1,v2,bottom)	)
			{
				//  
				Raster_Line(v1.x,v1.y,v2.x,v2.y,edge.color_id);
			}
		}
	}
}

//---------------------------------------------------------------------------
//      
void Draw_Scene_Scan()
{
	//         
	Init_Raster_Frame();
	//      ( )
	Plane_struct plane(0,0,-1,-MINIMUM_Z);
	//     
	Line_struct left(-1,0,0),right(1,0,RWidth);
	Line_struct top(0,-1,0),bottom(0,1,RHeight);
	//      
	for(long i=0;i<Number_Faces;i++)
	{
		const Face_struct &face=Faces[i];
		//    
		Circle_List_Template<Vertex3D_struct> Poly3D;
		Poly3D.append(new Vertex3D_struct(Vertices3D[face.vertex[0]]));
		Poly3D.append(new Vertex3D_struct(Vertices3D[face.vertex[1]]));
		Poly3D.append(new Vertex3D_struct(Vertices3D[face.vertex[2]]));
		//    
		if (TruncPolygon3D(Poly3D,plane))
		{
			//    
			Circle_List_Template<Vertex2D_struct> Poly2D;
			ProjectPolygon(Poly3D,Poly2D);

			//    
			if (	TruncPolygon2D(Poly2D,left)	&&	TruncPolygon2D(Poly2D,right)
				&&	TruncPolygon2D(Poly2D,top)	&&	TruncPolygon2D(Poly2D,bottom)	)
			{
				//     
				long ys,ye;
				Prepare_Raster_Polygon(Poly2D,i,ys,ye);
			}
			//     
			Poly2D.clear();
		}
		//     
		Poly3D.clear();
	}

	//  ,   
	Circle_List_Template<Edge_List_struct> Edges;
	//    
	for(long y=0;y<RHeight;y++)
	{
		//     
		for(Circle_List_Iterator_Template<Edge_List_struct> i(Edges);!i.end();)
		{
			Edge_List_struct *edge=i.get();
			//,   
			if (y>=edge->end)
			{
				delete i.remove();
			}
			else
			{
				//    
				edge->x+=edge->dx;
				edge->w+=edge->dw;
				i.step();
			}
		}

		//   
		while(LineEdges[y])
		{
			Edges.append(LineEdges[y]);
			LineEdges[y]=LineEdges[y]->next;
		}

		//    
		if (!Edges.start()) continue;

		//   ,    
		for(Circle_List_Iterator_Template<Edge_List_struct> e1(Edges);!e1.end();e1.step())
		{
			bool found=false;
			Circle_List_Iterator_Template<Edge_List_struct> e2(e1);
			for(e2.step();!e2.end();e2.step())
				if (e2.get()->ID==e1.get()->ID)	//     
				{
					//   e2   e1
					Edges.insert(Edges.remove(e2.get_node()),e1.get_node());
					found=true;
					break;
				}
			assert(found,"    ");
			e1.step();
		}

		//       
		for(Circle_List_Iterator_Template<Edge_List_struct> s(Edges);!s.end();)
		{
			Edge_List_struct *Edge1=s.get();	s.step();
			Edge_List_struct *Edge2=s.get();	s.step();
			//   
			if (Edge1->ID!=Edge2->ID)	break;

			// Edge1      Edge2
			if (Edge1->x>Edge2->x)	swap(Edge1,Edge2);
			//    
			long xs=(long)floor(Edge1->x+0.5);
			long xe=(long)floor(Edge2->x+0.5);
			//      
			if (xs<0)	xs=0;
			if (xe>RWidth)	xe=RWidth;
			//   
			if (xs>=xe)	continue;

			// 
			float dw=(Edge2->w-Edge1->w)/(Edge2->x-Edge1->x);
			float w=Edge1->w+dw*(xs+0.5f-Edge1->x);
			for(long x=xs;x<xe;x++)
			{
				Put_Pixel(x,y,Faces[Edge1->ID].color_id);
				// w-   
				w+=dw;
			}
		}
	}
	//  
	Edges.clear();
}

//---------------------------------------------------------------------------
//      
void Draw_Scene_RayTrace()
{
	//     
	Init_Raster_Frame();
              
	//     
	for (long y = 0; y != RHeight; ++y)
	{
		//      
		for (long x = 0; x != RWidth; ++x)
		{
			// ,        
			short color_id = 0;

			// ,         
			float currentW = 1e+10;


		//   
		for(long i=0;i<Number_Faces;i++)

		{
			const Face_struct &face=Faces[i];
			//     
			Vertex3D_struct vertex1(Vertices3D[face.vertex[0]]);
    		Vertex3D_struct vertex2(Vertices3D[face.vertex[1]]);
			Vertex3D_struct vertex3(Vertices3D[face.vertex[2]]);

			//       
    		float newW = CheckRayTriangle(Vector3D(0,0,0), PixelDirection(x,y),
										  Vector3D(vertex1.x, vertex1.y, vertex1.z),
										  Vector3D(vertex2.x, vertex2.y, vertex2.z),
										  Vector3D(vertex3.x, vertex3.y, vertex3.z));

                if(newW>-1.0)
			
				if (newW < currentW)
				{
					//    
					currentW = newW;
					//      
					color_id = face.color_id;
	
				}
		}

			//        
			Put_Pixel(x, y, color_id);
		}
	}


}



//---------------------------------------------------------------------------
//   
void Draw_Scene_ZBuffer();
void Draw_Scene_ScanZBuffer();
void Draw_Scene_ScanSBuffer();
void Draw_Scene_Roberts();
void Draw_Scene_RayTrace();
void Draw_Scene_Warnock();
void Draw_Scene_Painter();

//---------------------------------------------------------------------------
//   
void User_Test()
{
	//  scene_X, X -    
	if (!LoadScene("Data\\scene_2",2))	return;

	//   
	Set_Resolution_Factor(0);

	//  
	switch(7)
	{
		case 0:	Draw_Scene_Wire();			break;
		case 1:	Draw_Scene_Solid();			break;
		case 2:	Draw_Scene_Scan();			break;
	//	case 3:	Draw_Scene_ZBuffer();		break;
	//	case 4:	Draw_Scene_ScanZBuffer();	break;
	//	case 5:	Draw_Scene_ScanSBuffer();	break;
	//	case 6:	Draw_Scene_Roberts();		break;
		case 7:	Draw_Scene_RayTrace();		break;
	//	case 8:	Draw_Scene_Warnock();		break;
	//	case 9:	Draw_Scene_Painter();		break;
	}

	SaveBMP("result.bmp",ImageArray,Colors,RWidth,RHeight);

	//    
	Update_View(true);

	//   
	Clear_Scene();
}

//---------------------------------------------------------------------------
// -  
#define NUMBER_SCENES		5
// -   
#define NUMBER_RESOLUTIONS	5

//---------------------------------------------------------------------------
//  
void User_Main()
{
	//   
//	User_Test();	return;

	//      
	double time[NUMBER_SCENES][NUMBER_RESOLUTIONS];
	// -    
	long number_pixels[NUMBER_RESOLUTIONS];
	// -    
	long number_polygones[NUMBER_SCENES];

	for(int i=0;i<NUMBER_SCENES;i++)
	{
		//  scene_X, X -    
		if (!LoadScene("Data\\scene_2",i+1))	return;
		number_polygones[i]=Number_Faces;

		for(int j=0;j<NUMBER_RESOLUTIONS;j++)
		{
			//   
			Set_Resolution_Factor(NUMBER_RESOLUTIONS+1-j);
			number_pixels[j]=RWidth*RHeight;

			double takts=Get_Takts();

			//  
			switch(7)
			{
				case 0:	Draw_Scene_Wire();			break;
				case 1:	Draw_Scene_Solid();			break;
				case 2:	Draw_Scene_Scan();			break;
			//	case 3:	Draw_Scene_ZBuffer();		break;
			//	case 4:	Draw_Scene_ScanZBuffer();	break;
			//	case 5:	Draw_Scene_ScanSBuffer();	break;
			//	case 6:	Draw_Scene_Roberts();		break;
				case 7:	Draw_Scene_RayTrace();		break;
			//	case 8:	Draw_Scene_Warnock();		break;
			//	case 9:	Draw_Scene_Painter();		break;
			}

			//    
			time[i][j]=Get_Takts()-takts; 
			// Get_Takts     

			//    
			Update_View(true);
		}

		//       
		if (i==(NUMBER_SCENES-1))	SaveBMP("result.bmp",ImageArray,Colors,RWidth,RHeight);

		//   
		Clear_Scene();
	}

	//   
	FILE *file = fopen("results.txt","wt");
	assert(file,"   ");
	if (file)
	{
		fprintf(file,"Pixels\\Polygones ");
		for(int i=0;i<NUMBER_SCENES;i++)
			fprintf(file,"    %5d",number_polygones[i]);
		fprintf(file,"\n");

		for(int j=0;j<NUMBER_RESOLUTIONS;j++)
		{
			fprintf(file," %5d           ",number_pixels[j]);
			for(int i=0;i<NUMBER_SCENES;i++)
			{
				fprintf(file," %8d",long(time[i][j]/1000));
			}
			fprintf(file,"\n");
		}
	}

	fclose(file);
}

